/*

Copyleft (L) 2003-2004  AlterWare Beast Inc.   >;->

Module Name:        PciDump.c
Abstract:           This driver allows you to access PCI bus directly from UserMode.
Authors:            Alter
Environment:        kernel mode only

 */

#include "PCIDump.h"

PDRIVER_OBJECT DriverObject;
PDEVICE_OBJECT PciDumpDevice = NULL;

UNICODE_STRING      unicodeDeviceNameDos;

CHAR Copyright[] = "PciDumpEx driver v0.1e Copyright (c) 2003-2004 by Alexander A. Telyatnikov\n"
                   "home site: http://www.alter.org.ua\n";
/*

    Arguments:
    
    DriverObject - The PCI Dump driver object.

    Returns:    NTSTATUS
 */

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT driverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
    CCHAR		deviceNameBuffer[MAXIMUM_FILENAME_LENGTH];
    CCHAR		deviceNameBufferDos[MAXIMUM_FILENAME_LENGTH];
    ANSI_STRING         deviceName;
    ANSI_STRING         deviceNameDos;
    UNICODE_STRING      unicodeDeviceName;
    ULONG               i;
    NTSTATUS            status;

    DriverObject = driverObject;

    for(i=0; i<=IRP_MJ_MAXIMUM_FUNCTION; i++)
        DriverObject->MajorFunction[i] = NULL;

    // Set up the device driver entry points.
    DriverObject->MajorFunction[IRP_MJ_CREATE] = PciDumpCreateClose;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP] = PciDumpCreateClose;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = PciDumpCreateClose;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = PciDumpDeviceControl;

    DriverObject->DriverUnload = PciDumpUnload;

    sprintf((PCHAR)&deviceNameBuffer, NT_PciDump_NAME);
    KdPrint(("PciDumpInitialize: Checking \\Device\\PciDump\n"));
    RtlInitAnsiString(&deviceName, (PCHAR)&deviceNameBuffer);
    RtlAnsiStringToUnicodeString(&unicodeDeviceName, &deviceName, TRUE);
    // Create the device object for PciDump.
    status = IoCreateDevice(DriverObject, 0 /*sizeof(LM_DEVICE_EXTENSION)*/, &unicodeDeviceName,
                        FILE_DEVICE_UNKNOWN, NULL, FALSE, &PciDumpDevice);
    if(!NT_SUCCESS(status)) {
        RtlFreeUnicodeString(&unicodeDeviceName);
        return status;
    }

    sprintf((PCHAR)&deviceNameBufferDos, NT_PciDump_DOS_NAME);
    KdPrint(("PciDumpInitialize: Checking \\DosDevices\\PciDump\n"));
    RtlInitAnsiString(&deviceNameDos, (PCHAR)&deviceNameBufferDos);
    RtlAnsiStringToUnicodeString(&unicodeDeviceNameDos, &deviceNameDos, TRUE);
    status = IoCreateSymbolicLink (&unicodeDeviceNameDos, &unicodeDeviceName);
    if(!NT_SUCCESS(status)) {
        IoDeleteDevice(PciDumpDevice);
        RtlFreeUnicodeString(&unicodeDeviceName);
        RtlFreeUnicodeString(&unicodeDeviceNameDos);
        return status;
    }
    RtlFreeUnicodeString(&unicodeDeviceName);

    return STATUS_SUCCESS;

} // DriverEntry()

/*
    This routine services open commands. It establishes
    the driver's existance by returning status success.

    Arguments:

    DeviceObject - Context for the activity.
    Irp          - The device control argument block.

    Returns:    NT Status
 */
NTSTATUS
PciDumpCreateClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    return STATUS_SUCCESS;
} // end PciDumpCreateClose()


/*
    This device control dispatcher handles only the disk performance
    device control. All others are passed down to the disk drivers.
    The disk performane device control returns a current snapshot of
    the performance data.

    Arguments:

    DeviceObject - Context for the activity.
    Irp          - The device control argument block.

    Returns:    NT Status
 */
NTSTATUS
PciDumpDeviceControl(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp
    )
{
    KdPrint(("PciDumpDeviceControl\n"));
//    PLM_DEVICE_EXTENSION deviceExtension = (PLM_DEVICE_EXTENSION)DeviceObject->DeviceExtension;
    PIO_STACK_LOCATION  irpStack = IoGetCurrentIrpStackLocation(Irp);
    ULONG               ioctlCode;
    NTSTATUS            status;

    Irp->IoStatus.Information = 0;
    ioctlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;
    switch (ioctlCode) {
    case IOCTL_PciDump_GetHeader: {
        KdPrint(("PciDump: GetHeader\n"));
        // We have <DeviceName> & its attributes.
        PPciDump_GetHeader_USER_IN  buffer;
        PPciDump_GetHeader_USER_OUT obuffer;
        PCI_SLOT_NUMBER slotData;

        ULONG length = sizeof(PciDump_GetHeader_USER_IN);
        CHECK_IN_BUFFER();
        length = sizeof(PciDump_GetHeader_USER_OUT); 
        CHECK_OUT_BUFFER();
        length -= sizeof(ULONG);

        buffer = (PPciDump_GetHeader_USER_IN)(Irp->AssociatedIrp.SystemBuffer);
        obuffer = (PPciDump_GetHeader_USER_OUT)(Irp->AssociatedIrp.SystemBuffer);

        slotData.u.AsULONG = 0;
        slotData.u.bits.DeviceNumber   = buffer->DeviceNumber;
        slotData.u.bits.FunctionNumber = buffer->FunctionNumber;
        obuffer->status =
            HalGetBusData(PCIConfiguration,
                          buffer->BusNumber,
                          slotData.u.AsULONG, 
                          &(obuffer->pciData),
                          length);

        KdPrint(("PciDumpInitialize: obuffer->status = %x\n", obuffer->status));
        Irp->IoStatus.Information = length+sizeof(ULONG);
        status = STATUS_SUCCESS;
        break;
    }
    case IOCTL_PciDump_GetDataByOffset: {
        KdPrint(("PciDump: GetDataByOffset\n"));
        // We have <DeviceName> & its attributes.
        PPciDump_GetDataByOffset_USER_IN  buffer;
        PPciDump_GetDataByOffset_USER_OUT obuffer;
        PCI_SLOT_NUMBER slotData;

        ULONG length = sizeof(PciDump_GetDataByOffset_USER_IN);
        CHECK_IN_BUFFER();
        buffer = (PPciDump_GetDataByOffset_USER_IN)(Irp->AssociatedIrp.SystemBuffer);
        length = sizeof(PciDump_GetDataByOffset_USER_OUT)+buffer->Length; 
        CHECK_OUT_BUFFER();
        obuffer = (PPciDump_GetDataByOffset_USER_OUT)(Irp->AssociatedIrp.SystemBuffer);

        slotData.u.AsULONG = 0;
        slotData.u.bits.DeviceNumber   = buffer->DeviceNumber;
        slotData.u.bits.FunctionNumber = buffer->FunctionNumber;
        obuffer->status =
            HalGetBusDataByOffset(PCIConfiguration,
                          buffer->BusNumber,
                          slotData.u.AsULONG, 
                          obuffer+1,
                          buffer->Offset,
                          buffer->Length);

        KdPrint(("PciDumpInitialize: obuffer->status = %x\n", obuffer->status));
        Irp->IoStatus.Information = obuffer->status;
        status = STATUS_SUCCESS;
        break;
    }
    default:

        status = STATUS_INVALID_DEVICE_REQUEST;
        break;
    }

PciDumpSpecDevCtrlCompleteRequest:

    Irp->IoStatus.Status = status;    // Update IRP with completion status.
    IoCompleteRequest(Irp, IO_DISK_INCREMENT);    // Complete the request.
    KdPrint(( "PciDumpDeviceControl: Status is %lx\n", status));
    return status;

} // end PciDumpDeviceControl()


/*
    This routine services driver unload commands.

    Arguments:

    DeviceObject - Context for the activity.

    Returns:    NT Status
 */
VOID
PciDumpUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    IoDeleteSymbolicLink(&unicodeDeviceNameDos);
    
    RtlFreeUnicodeString(&unicodeDeviceNameDos);

    if(PciDumpDevice) {
        IoDeleteDevice(PciDumpDevice);
        PciDumpDevice = NULL;
    }
    return;
} // end PciDumpCreateClose()

